// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/20AO8C500K/driver/device.c $
// $Rev: 53736 $
// $Date: 2023-09-21 16:09:42 -0500 (Thu, 21 Sep 2023) $

// 20AO8C500K: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model		Vendor	Device	SubVen	SubDev	type
	{ "20AO8C500K",	0x10B5, 0x9056, 0x10B5, 0x3574,	GSC_DEV_TYPE_20AO8C500K	},
	{ NULL }
};



//*****************************************************************************
static void _channels_qty_compute(dev_data_t* dev)
{
	u32	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 17, 16);

	dev->cache.channels_max	= 8;

	switch (field)
	{
		default:
		case 2:
		case 3:		printf("%s: unrecognized Channel Count option: %d\n",
							dev->model,
							(int) field);
					printf("%s: defaulting to 8 channels\n", dev->model);

					FALLTHROUGH;	/* fall through */

		case 0:		dev->cache.channel_qty	= 8;
					break;

		case 1:		dev->cache.channel_qty	= 4;
					break;
	}
}



//*****************************************************************************
static void _fref_default_compute(dev_data_t* dev)
{
	u32	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 19, 18);

	switch (field)
	{
		default:
		case 1:
		case 2:
		case 3:		printf("%s: unrecognized Master Clock Frequency option: %d\n",
							dev->model,
							(int) field);
					printf("%s: defaulting to 49.152 MHz\n", dev->model);

					FALLTHROUGH;	/* fall through */

		case 0:		dev->cache.fref_default	= 49152000;	break;
	}
}



//*****************************************************************************
static void _output_disconnect_compute(dev_data_t* dev)
{
	u32	field	= dev->cache.gsc_acr_32 & D22;

	dev->cache.output_disconnect	= field ? 0 : 1;
}



//*****************************************************************************
static void _output_filter_compute(dev_data_t* dev)
{
	u32	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 25, 24);

	switch (field)
	{
		default:
		case 1:
		case 2:
		case 3:		printf("%s: unrecognized Output Filter option: %d\n",
							dev->model,
							(int) field);
					printf("%s: defaulting to 100KHz\n", dev->model);

					FALLTHROUGH;	/* fall through */

		case 0:		dev->cache.output_filter	= 100000;	break;
	}
}



//*****************************************************************************
static void _output_range_compute(dev_data_t* dev)
{
	u32	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 21, 20);

	switch (field)
	{
		default:
		case 2:
		case 3:		printf("%s: unrecognized Output Range option: %d\n",
							dev->model,
							(int) field);
					printf("%s: defaulting to +-10\n", dev->model);

					FALLTHROUGH;	/* fall through */

		case 0:	dev->cache.output_range	= AO8C500K_QUERY_OUTPUT_RANGE_10;	break;
		case 1:	dev->cache.output_range	= AO8C500K_QUERY_OUTPUT_RANGE_5;	break;
	}
}



//*****************************************************************************
static void _pin_a33_compute(dev_data_t* dev)
{
	u32	field	= dev->cache.gsc_acr_32 & D23;

	dev->cache.pin_a33	= field
						? AO8C500K_QUERY_PIN_A33_RMT_GND_S
						: AO8C500K_QUERY_PIN_A33_OUT_RTN;
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(AO8C500K_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_acr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_ACR));
		dev->vaddr.gsc_aobr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_AOBR));
		dev->vaddr.gsc_asiocr_32	= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_ASIOCR));
		dev->vaddr.gsc_bcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_BCR));
		dev->vaddr.gsc_boor_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_BOOR));
		dev->vaddr.gsc_diopr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_DIOPR));
		dev->vaddr.gsc_obsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_OBSR));
		dev->vaddr.gsc_obtr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_OBTR));
		dev->vaddr.gsc_ocr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_OCR));
		dev->vaddr.gsc_psr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_PSR));
		dev->vaddr.gsc_ragr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_RAGR));
		dev->vaddr.gsc_rbgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO8C500K_GSC_RBGR));

		// Data cache initialization.
		dev->cache.gsc_acr_32		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_acr_32);

		#define	NDIV_MAX			0xFFFFFF
		#define	NDIV_MIN			98

		dev->cache.autocal_ms		= 6000;
		dev->cache.fifo_size		= _256K;
		dev->cache.initialize_ms	= 3;
		dev->cache.ndiv_max			= NDIV_MAX;
		dev->cache.ndiv_min			= NDIV_MIN;

		_channels_qty_compute(dev);
		_fref_default_compute(dev);
		_output_disconnect_compute(dev);
		_output_filter_compute(dev);
		_output_range_compute(dev);
		_pin_a33_compute(dev);

		dev->cache.fgen_min			= dev->cache.fref_default / NDIV_MAX;
		dev->cache.fgen_max			= dev->cache.fref_default / NDIV_MIN;
		dev->cache.fsamp_min		= dev->cache.fgen_min;
		dev->cache.fsamp_max		= dev->cache.fgen_max;

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_WRITE
			| GSC_DMA_CAP_DMDMA_WRITE;
		ret	= gsc_dma_create(dev, dma, dma);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


