// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/20AO8C500K/driver/ioctl.c $
// $Rev: 53736 $
// $Date: 2023-09-21 16:09:42 -0500 (Thu, 21 Sep 2023) $

// 20AO8C500K: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTOCAL_START			D28			// BCR
#define	AUTOCAL_PASS			D29			// BCR
#define	INIT_START				D31			// BCR

#define	IRQ_INIT_DONE			0x0001		// PSR
#define	IRQ_AUTOCAL_DONE		0x0002		// PSR
#define	IRQ_MASK				0xFFFF		// PSR
#define	IRQ_ACTIVE				0xFFFF		// PSR



//*****************************************************************************
static void _psr_mod(dev_data_t* dev, u32 value, u32 mask)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_psr_32, value & 0xFFFF, mask | 0xFFFF0000);
}



//*****************************************************************************
static int _query(dev_data_t* dev, s32* arg)
{
	int		ret	= 0;

	switch (arg[0])
	{
		default:	arg[0]	= AO8C500K_IOCTL_QUERY_ERROR;
					ret		= -EINVAL;
					break;

		case AO8C500K_QUERY_AUTOCAL_MS:			arg[0]	= dev->cache.autocal_ms;		break;
		case AO8C500K_QUERY_CHANNEL_MAX:		arg[0]	= dev->cache.channels_max;		break;
		case AO8C500K_QUERY_CHANNEL_QTY:		arg[0]	= dev->cache.channel_qty;		break;
		case AO8C500K_QUERY_COUNT:				arg[0]	= AO8C500K_QUERY_LAST;			break;
		case AO8C500K_QUERY_DEVICE_TYPE:		arg[0]	= dev->board_type;				break;
		case AO8C500K_QUERY_FGEN_MAX:			arg[0]	= dev->cache.fgen_max;			break;
		case AO8C500K_QUERY_FGEN_MIN:			arg[0]	= dev->cache.fgen_min;			break;
		case AO8C500K_QUERY_FIFO_SIZE:			arg[0]	= dev->cache.fifo_size;			break;
		case AO8C500K_QUERY_FREF_DEFAULT:		arg[0]	= dev->cache.fref_default;		break;
		case AO8C500K_QUERY_FSAMP_MAX:			arg[0]	= dev->cache.fsamp_max;			break;
		case AO8C500K_QUERY_FSAMP_MIN:			arg[0]	= dev->cache.fsamp_min;			break;
		case AO8C500K_QUERY_INIT_MS:			arg[0]	= dev->cache.initialize_ms;		break;
		case AO8C500K_QUERY_NDIV_MAX:			arg[0]	= dev->cache.ndiv_max;			break;
		case AO8C500K_QUERY_NDIV_MIN:			arg[0]	= dev->cache.ndiv_min;			break;
		case AO8C500K_QUERY_OUTPUT_DISCONNECT:	arg[0]	= dev->cache.output_disconnect;	break;
		case AO8C500K_QUERY_OUTPUT_FILTER:		arg[0]	= dev->cache.output_filter;		break;
		case AO8C500K_QUERY_OUTPUT_RANGE:		arg[0]	= dev->cache.output_range;		break;
		case AO8C500K_QUERY_PIN_A33:			arg[0]	= dev->cache.pin_a33;			break;
	}

	return(ret);
}



//*****************************************************************************
static int _init_start(dev_data_t* dev, void* arg)
{
	// Initiate initialization.
	os_reg_mem_tx_u32(NULL, dev->vaddr.gsc_bcr_32, INIT_START);
	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	int			i;
	int			mask;
	long		ms			= dev->cache.initialize_ms + 5000;
	long		ms_total	= ms;
	u32			reg;
	int			ret			= 0;
	os_sem_t	sem;
	int			tmp;
	VADDR_T		va			= dev->vaddr.gsc_bcr_32;
	int			value;
	gsc_wait_t	wt;

	if ((dev->irq.opened) && (gsc_global.driver_unloading == 0))
	{
		ms_total	*= 2;

		// Safely select the Initialize Done interrupt.
		mask	= IRQ_MASK | IRQ_ACTIVE;
		value	= IRQ_INIT_DONE;
		_psr_mod(dev, value, mask);

		// Wait for the local interrupt.
		os_sem_create(&sem);	// dummy, required for wait operations.
		memset(&wt, 0, sizeof(wt));
		wt.flags		= GSC_WAIT_FLAG_INTERNAL;
		wt.gsc			= AO8C500K_WAIT_GSC_INIT_DONE;
		wt.timeout_ms	= ms;
		ret				= gsc_wait_event(dev, &wt, _init_start, NULL, &sem);
		os_sem_destroy(&sem);

		if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
		{
			ret	= ret ? ret : -ETIMEDOUT;
			printf(	"%s: INITILIZE DONE IRQ TIMED OUT AFTER %ld ms.\n",
					dev->model,
					ms);
		}
	}
	else
	{
		_init_start(dev, NULL);
	}

	// Manually wait for completion as the IRQ wait may have ended early.
	tmp	= gsc_poll_u32(dev, ms, va, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	reg	= os_reg_mem_rx_u32(dev, va);

	if (reg & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	return(ret);
}



//*****************************************************************************
static int _autocal_start(dev_data_t* dev, void* arg)
{
	// Initiate autocalibration.
	os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bcr_32, AUTOCAL_START, AUTOCAL_START);
	return(0);
}



//*****************************************************************************
static int _autocal(dev_data_t* dev, void* arg)
{
	u32			mask;
	long		ms		= dev->cache.autocal_ms + 5000;
	u32			reg;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	VADDR_T		va		= dev->vaddr.gsc_bcr_32;
	u32			value;
	gsc_wait_t	wt;

	// Safely select the Autocal Done interrupt.
	mask	= IRQ_AUTOCAL_DONE;
	value	= IRQ_AUTOCAL_DONE;
	_psr_mod(dev, value, mask);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= AO8C500K_WAIT_GSC_AUTOCAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _autocal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE DONE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, va, AUTOCAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	reg	= os_reg_mem_rx_u32(dev, va);

	if (reg & AUTOCAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	reg	= os_reg_mem_rx_u32(dev, va);

	if ((reg & AUTOCAL_PASS) == 0)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				ms);
	}

	return(ret);
}



//*****************************************************************************
static int _autocal_status(dev_data_t* dev, s32* arg)
{
	u32	bcr;

	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & D28)
		arg[0]	= AO8C500K_AUTOCAL_STATUS_ACTIVE;
	else if (bcr & D29)
		arg[0]	= AO8C500K_AUTOCAL_STATUS_PASS;
	else
		arg[0]	= AO8C500K_AUTOCAL_STATUS_FAIL;

	return(0);
}



//*****************************************************************************
static int _ao_config(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_AO_CONFIG_SE,
		AO8C500K_AO_CONFIG_DIFF,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _aux_clock_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_AUX_MODE_DISABLE,
		AO8C500K_AUX_MODE_INPUT,
		AO8C500K_AUX_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 1, 0);
	return(ret);
}



//*****************************************************************************
static int _aux_in_pol(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_AUX_POL_L2H,
		AO8C500K_AUX_POL_H2L,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _aux_out_pol(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_AUX_POL_L2H,
		AO8C500K_AUX_POL_H2L,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 9, 9);
	return(ret);
}



//*****************************************************************************
static int _aux_noise(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_AUX_NOISE_LOW,
		AO8C500K_AUX_NOISE_HIGH,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _aux_trigger_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_AUX_MODE_DISABLE,
		AO8C500K_AUX_MODE_INPUT,
		AO8C500K_AUX_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 3, 2);
	return(ret);
}



//*****************************************************************************
static int _buf_clear(dev_data_t* dev, s32* arg)
{
	#define	CLEAR		D11		// BOOR
	#define	DATA		D16		// BOOR
	#define	FRAME		D17		// BOOR

	int		i;
	u32		reg;
	int		ret	= 0;
	VADDR_T	va	= dev->vaddr.gsc_boor_32;

	// Clear the buffer.
	os_reg_mem_mx_u32(dev, va, CLEAR, CLEAR);

	// Wait for the bit to clear.

	for (i = 0;; i++)
	{
		reg	= os_reg_mem_rx_u32(dev, va);

		if ((reg & CLEAR) == 0)
			break;

		if (i >= 250)
		{
			ret	= -EINVAL;
			printf(	"%s: The analog output buffer took too long to clear.\n",
					DEV_NAME);
			break;
		}
	}

	// Clear the Overflow status bit.
	os_reg_mem_mx_u32(dev, va, 0, DATA);

	// Clear the Frame Overflow status bit.
	os_reg_mem_mx_u32(dev, va, 0, FRAME);

	return(ret);
}



//*****************************************************************************
static int _buf_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_BUF_MODE_OPEN,
		AO8C500K_BUF_MODE_CIRC,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _buf_over_data(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_BUF_OVER_DATA_NO,
		AO8C500K_BUF_OVER_DATA_YES,
		-1	// terminate list
	};

	int	ret;

	if ((arg[0] == AO8C500K_BUF_OVER_DATA_CLR) ||
		(arg[0] == AO8C500K_BUF_OVER_DATA_CHK))
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 16, 16);
	}
	else
	{
		ret	= -EINVAL;
	}

	return(ret);
}



//*****************************************************************************
static int _buf_over_frame(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_BUF_OVER_FRAME_NO,
		AO8C500K_BUF_OVER_FRAME_YES,
		-1	// terminate list
	};

	int	ret;

	if ((arg[0] == AO8C500K_BUF_OVER_FRAME_CLR) ||
		(arg[0] == AO8C500K_BUF_OVER_FRAME_CHK))
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 17, 17);
	}
	else
	{
		ret	= -EINVAL;
	}

	return(ret);
}



//*****************************************************************************
static int _buf_status(dev_data_t* dev, s32* arg)
{
	u32	boor;

	boor	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_boor_32);

	if (boor & D12)
		arg[0]	= AO8C500K_BUF_STATUS_EMPTY;
	else if (boor & D15)
		arg[0]	= AO8C500K_BUF_STATUS_FULL;
	else if (boor & D13)
		arg[0]	= AO8C500K_BUF_STATUS_AF;
	else
		arg[0]	= AO8C500K_BUF_STATUS_AE;

	return(0);
}



//*****************************************************************************
static int _buf_fill_level(dev_data_t* dev, s32* arg)
{
	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_range_reg(dev, arg, 0, _256K, dev->vaddr.gsc_obsr_32, 18, 0);
	return(ret);
}



//*****************************************************************************
static int _buf_thresh_level(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, _256K, dev->vaddr.gsc_obtr_32, 18, 0);
	return(ret);
}



//*****************************************************************************
static int _buf_thresh_status(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_BUF_THRESH_STATUS_CLR,
		AO8C500K_BUF_THRESH_STATUS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_obtr_32, 20, 20);
	return(ret);
}



//*****************************************************************************
static int _burst_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_BURST_ENABLE_NO,
		AO8C500K_BURST_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _burst_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_BURST_READY_NO,
		AO8C500K_BURST_READY_YES,
		-1	// terminate list
	};

	int		ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _burst_trig_src(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_BURST_TRIG_SRC_RBG,
		AO8C500K_BURST_TRIG_SRC_EXT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 15, 14);
	return(ret);
}



//*****************************************************************************
static int _channel_sel(dev_data_t* dev, s32* arg)
{
	int	ret;

	if (dev->cache.channel_qty == 8)
		ret	= gsc_s32_range_reg(dev, arg, 0, 0xFF, dev->vaddr.gsc_ocr_32, 7, 0);
	else
		ret	= gsc_s32_range_reg(dev, arg, 0, 0x0F, dev->vaddr.gsc_ocr_32, 3, 0);

	return(ret);
}



//*****************************************************************************
static int _clock_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_CLOCK_ENABLE_NO,
		AO8C500K_CLOCK_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 5, 5);
	return(ret);
}



//*****************************************************************************
static int _clock_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_CLOCK_READY_NO,
		AO8C500K_CLOCK_READY_YES,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 6, 6);
	return(ret);
}



//*****************************************************************************
static int _clock_src(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_CLOCK_SRC_RAG,
		AO8C500K_CLOCK_SRC_EXT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 13, 12);
	return(ret);
}



//*****************************************************************************
static int _clock_sw(dev_data_t* dev, s32* arg)
{
	int	ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D7, D7);
	ret	= gsc_poll_u32(dev, dev->io.tx.timeout_s * 1000, dev->vaddr.gsc_boor_32, D7, 0);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_DATA_FORMAT_2S_COMP,
		AO8C500K_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 25, 25);
	return(ret);
}



//*****************************************************************************
static int _dio_dir_out(dev_data_t* dev, s32* arg)
{
	u32	diopr;
	int	ret	= 0;

	switch (arg[0])
	{
		default:

			ret	= -EINVAL;
			break;

		case -1:

			diopr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_diopr_32);
			arg[0]	= diopr & (D15 | D7);
			break;

		case AO8C500K_DIO_DIR_OUT_NONE:
		case AO8C500K_DIO_DIR_OUT_D3_D0:
		case AO8C500K_DIO_DIR_OUT_D7_D4:
		case AO8C500K_DIO_DIR_OUT_D7_D0:

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_diopr_32, arg[0], D15 | D7);
			break;
	}

	return(ret);
}



//*****************************************************************************
static int _dio_read(dev_data_t* dev, s32* arg)
{
	u32	diopr;

	diopr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_diopr_32);
	arg[0]	= ((diopr & 0xF00) >> 4) | (diopr & 0xF);
	return(0);
}



//*****************************************************************************
static int _dio_write(dev_data_t* dev, s32* arg)
{
	int	ret	= 0;
	u32	v;

	if (arg[0] == -1)
	{
		ret	= _dio_read(dev, arg);
	}
	else if ((arg[0] < 0) || (arg[0] > 0xFF))
	{
		ret	= -EINVAL;
	}
	else
	{
		v	= ((arg[0] << 4) | arg[0])  & 0xF0F;
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_diopr_32, v, 0xF0F);
	}

	return(ret);
}



//*****************************************************************************
static int _gen_a_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_GEN_ENABLE_NO,
		AO8C500K_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 21, 21);
	return(ret);
}



//*****************************************************************************
static int _gen_a_ndiv(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.ndiv_min,
			dev->cache.ndiv_max,
			dev->vaddr.gsc_ragr_32,
			23, 0);
	return(ret);
}



//*****************************************************************************
static int _gen_b_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_GEN_ENABLE_NO,
		AO8C500K_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 22, 22);
	return(ret);
}



//*****************************************************************************
static int _gen_b_ndiv(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.ndiv_min,
			dev->cache.ndiv_max,
			dev->vaddr.gsc_rbgr_32,
			23, 0);
	return(ret);
}



//*****************************************************************************
static int _io_dma_threshold(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(arg, 0, 0xFFFFFFF, &dev->io.tx.dma_threshold);
	return(ret);
}



//*****************************************************************************
static int _io_pio_threshold(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(arg, 0, 0xFFFFFFF, &dev->io.tx.pio_threshold);
	return(ret);
}



//*****************************************************************************
static int _irq_enable(dev_data_t* dev, s32* arg)
{
	u32	psr;
	int	ret	= 0;

	if (arg[0] == -1)
	{
		// Retrieve the current setting.
		psr		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_psr_32);
		arg[0]	= psr & AO8C500K_IRQ_ALL;
	}
	else if ((arg[0] & AO8C500K_IRQ_ALL) != arg[0])
	{
		ret	= -EINVAL;
	}
	else
	{
		_psr_mod(dev, arg[0], AO8C500K_IRQ_ALL);
	}

	return(ret);
}



//*****************************************************************************
static int _load_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_LOAD_READY_NO,
		AO8C500K_LOAD_READY_YES,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _load_request(dev_data_t* dev, s32* arg)
{
	long	ms;
	int		ret;

	// Make the request.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D9, D9);

	// Manually wait for completion.
	ms	= dev->io.tx.timeout_s * 1000;
	ret	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_boor_32, D10, D10);
	return(ret);
}



//*****************************************************************************
static int _output_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_OUTPUT_ENABLE_NO,
		AO8C500K_OUTPUT_ENABLE_YES,
		-1	// terminate list
	};

	int	ret	= 0;

	if (dev->cache.output_disconnect)
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 7, 7);
	}
	else
	{
		// The outputs are always connected when the Output Disconnect
		// feature is not supported.

		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case -1:

				arg[0]	= AO8C500K_OUTPUT_ENABLE_YES;
				break;

			case AO8C500K_OUTPUT_ENABLE_NO:

				arg[0]	= -1;
				break;

			case AO8C500K_OUTPUT_ENABLE_YES:

				break;
		}
	}

	return(ret);
}



//*****************************************************************************
static int _output_filter(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_OUTPUT_FILTER_NO,
		AO8C500K_OUTPUT_FILTER_YES,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 24, 24);
	return(ret);
}



//*****************************************************************************
static int _output_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_OUTPUT_MODE_SEQUENTIAL,
		AO8C500K_OUTPUT_MODE_SIMULTANEOUS ,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 18, 18);
	return(ret);
}



//*****************************************************************************
// Low Range: +-5 and +-2.5
static int _range_5(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret;

	switch (arg[0])
	{
		default:

			ret	= -EINVAL;
			break;

		case -1:

			reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_ocr_32);
			arg[0]	= (reg & D8) ? AO8C500K_RANGE_5 : AO8C500K_RANGE_2_5;
			ret		= 0;
			break;

		case AO8C500K_RANGE_2_5:

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_ocr_32, 0, D8);
			arg[0]	= -1;
			ret		= _range_5(dev, arg);
			break;

		case AO8C500K_RANGE_5:

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_ocr_32, D8, D8);
			arg[0]	= -1;
			ret		= _range_5(dev, arg);
			break;
	}


	return(ret);
}



//*****************************************************************************
// Low Range: +-10 and +-5
static int _range_10(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret;

	switch (arg[0])
	{
		default:

			ret	= -EINVAL;
			break;

		case -1:

			reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_ocr_32);
			arg[0]	= (reg & D8) ? AO8C500K_RANGE_10 : AO8C500K_RANGE_5;
			ret		= 0;
			break;

		case AO8C500K_RANGE_5:

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_ocr_32, 0, D8);
			arg[0]	= -1;
			ret		= _range_10(dev, arg);
			break;

		case AO8C500K_RANGE_10:

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_ocr_32, D8, D8);
			arg[0]	= -1;
			ret		= _range_10(dev, arg);
			break;
	}

	return(ret);
}



//*****************************************************************************
static int _range(dev_data_t* dev, s32* arg)
{
	int	ret;

	if (dev->cache.output_range == AO8C500K_QUERY_OUTPUT_RANGE_5)
		ret	= _range_5(dev, arg);
	else
		ret	= _range_10(dev, arg);

	return(ret);
}



//*****************************************************************************
static int _remote_gnd_sense(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_REMOTE_GND_SENSE_NO,
		AO8C500K_REMOTE_GND_SENSE_YES,
		-1	// terminate list
	};

	int	ret	= 0;

	if (dev->cache.pin_a33 == AO8C500K_QUERY_PIN_A33_RMT_GND_S)
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 6, 6);
	}
	else
	{
		// The Remote Ground Sense feature is unsupported.

		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case -1:

				arg[0]	= AO8C500K_REMOTE_GND_SENSE_NO;
				break;

			case AO8C500K_REMOTE_GND_SENSE_NO:

				break;

			case AO8C500K_REMOTE_GND_SENSE_YES:

				arg[0]	= -1;
				break;
		}
	}

	return(ret);
}



//*****************************************************************************
static int _trigger_sw(dev_data_t* dev, s32* arg)
{
	int	ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D20, D20);
	ret	= gsc_poll_u32(dev, dev->io.tx.timeout_s * 1000, dev->vaddr.gsc_boor_32, D20, 0);
	return(ret);
}



//*****************************************************************************
static int _tx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_write_abort_active_xfer(dev, &dev->io.tx);
	return(0);
}



//*****************************************************************************
static int _tx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		GSC_IO_MODE_DMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.tx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _tx_io_over_data(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_TX_IO_OVERFLOW_CHECK,
		AO8C500K_TX_IO_OVERFLOW_IGNORE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.tx.over_data);
	return(ret);
}



//*****************************************************************************
static int _tx_io_over_frame(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_TX_IO_OVERFLOW_CHECK,
		AO8C500K_TX_IO_OVERFLOW_IGNORE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.tx.over_frame);
	return(ret);
}



//*****************************************************************************
static int _tx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AO8C500K_IO_TIMEOUT_MIN,
			AO8C500K_IO_TIMEOUT_INFINITE,
			&dev->io.tx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _xcvr_output(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_XCVR_OUTPUT_NO,
		AO8C500K_XCVR_OUTPUT_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 11, 11);
	return(ret);
}



//*****************************************************************************
static int _xcvr_pol(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_XCVR_POL_DEFAULT,
		AO8C500K_XCVR_POL_INVERT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 17, 17);
	return(ret);
}



//*****************************************************************************
static int _xcvr_type(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8C500K_XCVR_TYPE_TTL,
		AO8C500K_XCVR_TYPE_LVDS,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 5, 5);
	return(ret);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ AO8C500K_IOCTL_REG_READ,			(void*) gsc_reg_read_ioctl		},
	{ AO8C500K_IOCTL_REG_WRITE,			(void*) gsc_reg_write_ioctl		},
	{ AO8C500K_IOCTL_REG_MOD,			(void*) gsc_reg_mod_ioctl		},
	{ AO8C500K_IOCTL_QUERY,				(void*) _query					},
	{ AO8C500K_IOCTL_INITIALIZE,		(void*) initialize_ioctl		},
	{ AO8C500K_IOCTL_AUTOCAL,			(void*) _autocal				},
	{ AO8C500K_IOCTL_AUTOCAL_STATUS,	(void*) _autocal_status			},
	{ AO8C500K_IOCTL_AO_CONFIG,			(void*) _ao_config				},
	{ AO8C500K_IOCTL_AUX_CLOCK_MODE,	(void*) _aux_clock_mode			},
	{ AO8C500K_IOCTL_AUX_IN_POL,		(void*) _aux_in_pol				},
	{ AO8C500K_IOCTL_AUX_OUT_POL,		(void*) _aux_out_pol			},
	{ AO8C500K_IOCTL_AUX_NOISE,			(void*) _aux_noise				},
	{ AO8C500K_IOCTL_AUX_TRIGGER_MODE,	(void*) _aux_trigger_mode		},
	{ AO8C500K_IOCTL_BUF_CLEAR,			(void*) _buf_clear				},
	{ AO8C500K_IOCTL_BUF_MODE,			(void*) _buf_mode				},
	{ AO8C500K_IOCTL_BUF_OVER_DATA,		(void*) _buf_over_data			},
	{ AO8C500K_IOCTL_BUF_OVER_FRAME,	(void*) _buf_over_frame			},
	{ AO8C500K_IOCTL_BUF_STATUS,		(void*) _buf_status				},
	{ AO8C500K_IOCTL_BUF_FILL_LEVEL,	(void*) _buf_fill_level			},
	{ AO8C500K_IOCTL_BUF_THRESH_LEVEL,	(void*) _buf_thresh_level		},
	{ AO8C500K_IOCTL_BUF_THRESH_STATUS,	(void*) _buf_thresh_status		},
	{ AO8C500K_IOCTL_BURST_ENABLE,		(void*) _burst_enable			},
	{ AO8C500K_IOCTL_BURST_READY,		(void*) _burst_ready			},
	{ AO8C500K_IOCTL_BURST_TRIG_SRC,	(void*) _burst_trig_src			},
	{ AO8C500K_IOCTL_CHANNEL_SEL,		(void*) _channel_sel			},
	{ AO8C500K_IOCTL_CLOCK_ENABLE,		(void*) _clock_enable			},
	{ AO8C500K_IOCTL_CLOCK_READY,		(void*) _clock_ready			},
	{ AO8C500K_IOCTL_CLOCK_SRC,			(void*) _clock_src				},
	{ AO8C500K_IOCTL_CLOCK_SW,			(void*) _clock_sw				},
	{ AO8C500K_IOCTL_DATA_FORMAT,		(void*) _data_format			},
	{ AO8C500K_IOCTL_DIO_DIR_OUT,		(void*) _dio_dir_out			},
	{ AO8C500K_IOCTL_DIO_READ,			(void*) _dio_read				},
	{ AO8C500K_IOCTL_DIO_WRITE,			(void*) _dio_write				},
	{ AO8C500K_IOCTL_GEN_A_ENABLE,		(void*) _gen_a_enable			},
	{ AO8C500K_IOCTL_GEN_A_NDIV,		(void*) _gen_a_ndiv				},
	{ AO8C500K_IOCTL_GEN_B_ENABLE,		(void*) _gen_b_enable			},
	{ AO8C500K_IOCTL_GEN_B_NDIV,		(void*) _gen_b_ndiv				},
	{ AO8C500K_IOCTL_IO_DMA_THRESHOLD,	(void*) _io_dma_threshold		},
	{ AO8C500K_IOCTL_IO_PIO_THRESHOLD,	(void*) _io_pio_threshold		},
	{ AO8C500K_IOCTL_IRQ_ENABLE,		(void*) _irq_enable				},
	{ AO8C500K_IOCTL_LOAD_READY,		(void*) _load_ready				},
	{ AO8C500K_IOCTL_LOAD_REQUEST,		(void*) _load_request			},
	{ AO8C500K_IOCTL_OUTPUT_ENABLE,		(void*) _output_enable			},
	{ AO8C500K_IOCTL_OUTPUT_FILTER,		(void*) _output_filter			},
	{ AO8C500K_IOCTL_OUTPUT_MODE,		(void*) _output_mode			},
	{ AO8C500K_IOCTL_RANGE,				(void*) _range					},
	{ AO8C500K_IOCTL_REMOTE_GND_SENSE,	(void*) _remote_gnd_sense		},
	{ AO8C500K_IOCTL_TRIGGER_SW,		(void*) _trigger_sw				},
	{ AO8C500K_IOCTL_TX_IO_ABORT,		(void*) _tx_io_abort			},
	{ AO8C500K_IOCTL_TX_IO_MODE,		(void*) _tx_io_mode				},
	{ AO8C500K_IOCTL_TX_IO_OVER_DATA,	(void*) _tx_io_over_data		},
	{ AO8C500K_IOCTL_TX_IO_OVER_FRAME,	(void*) _tx_io_over_frame		},
	{ AO8C500K_IOCTL_TX_IO_TIMEOUT,		(void*) _tx_io_timeout			},
	{ AO8C500K_IOCTL_WAIT_CANCEL,		(void*) gsc_wait_cancel_ioctl	},
	{ AO8C500K_IOCTL_WAIT_EVENT,		(void*) gsc_wait_event_ioctl	},
	{ AO8C500K_IOCTL_WAIT_STATUS,		(void*) gsc_wait_status_ioctl	},
	{ AO8C500K_IOCTL_XCVR_OUTPUT,		(void*) _xcvr_output			},
	{ AO8C500K_IOCTL_XCVR_POL,			(void*) _xcvr_pol				},
	{ AO8C500K_IOCTL_XCVR_TYPE,			(void*) _xcvr_type				},
	{ -1, NULL }
};


