// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/20AO8C500K/utils/fsamp_ao.c $
// $Rev: 53736 $
// $Date: 2023-09-21 16:09:42 -0500 (Thu, 21 Sep 2023) $

// 20AO8C500K: Utilities: source file

#include "main.h"



// macros *********************************************************************

#define	ABS(v)					(((v) < 0) ? (-(v)) : (v))
#define	LIMIT_MAX(v,max)		(((v) > (max)) ? (max) : (v))
#define	LIMIT_MIN(v,min)		(((v) < (min)) ? (min) : (v))
#define	LIMIT_RANGE(v,min,max)	(LIMIT_MIN((LIMIT_MAX((v),(max))),(min)))



/******************************************************************************
*
*	Function:	ao8c500k_fsamp_ao_compute
*
*	Purpose:
*
*		Determine the best configuration for the specified output sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	The desired sample rate.
*
*		ndiv	The Ndiv value is put here. MUST NOT BE NULL!
*
*		rate	Report the sample rate achieved here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao8c500k_fsamp_ao_compute(int fd, int index, int verbose, s32 fsamp, s32* ndiv, double* rate)
{
	double	delta;
	double	delta_best	= FLT_MAX;
	int		errs		= 0;
	s32		fref;
	double	fsamp_best	= 0;
	double	fsamp_got;
	s32		fsamp_max;	// Queried value.
	s32		fsamp_min;	// Queried value.
	double	fsamp_want;
	s32		ndiv_a;
	s32		ndiv_best	= 0;
	s32		ndiv_max;	// Queried value.
	s32		ndiv_min;	// Queried value.

	if (verbose)
	{
		gsc_label_index("AO Sample Rate", index);
		gsc_label_long_comma(fsamp);
		printf(" S/S");
	}

	errs	+= ao8c500k_query(fd, -1, 0, AO8C500K_QUERY_FREF_DEFAULT, &fref);
	errs	+= ao8c500k_query(fd, -1, 0, AO8C500K_QUERY_FSAMP_MAX, &fsamp_max);
	errs	+= ao8c500k_query(fd, -1, 0, AO8C500K_QUERY_FSAMP_MIN, &fsamp_min);
	errs	+= ao8c500k_query(fd, -1, 0, AO8C500K_QUERY_NDIV_MAX, &ndiv_max);
	errs	+= ao8c500k_query(fd, -1, 0, AO8C500K_QUERY_NDIV_MIN, &ndiv_min);

	fsamp_want	= (double) LIMIT_RANGE(fsamp, fsamp_min, fsamp_max);
	ndiv_a		= fref / fsamp_want;

	ndiv_min	= (ndiv_a <= ndiv_min) ? ndiv_min : (ndiv_a  - 1);
	ndiv_max	= (ndiv_a >= ndiv_max) ? ndiv_max : (ndiv_a  + 1);

	for (ndiv_a = ndiv_min; ndiv_a <= ndiv_max; ndiv_a++)
	{
		fsamp_got	= ((double) fref) / ndiv_a;
		delta		= ABS(fsamp_got - fsamp_want);

		if (delta < delta_best)
		{
			delta_best	= delta;
			fsamp_best	= fsamp_got;
			ndiv_best	= ndiv_a;
		}
	}

	if (ndiv)
		ndiv[0]	= ndiv_best;

	if (rate)
		rate[0]	= fsamp_best;

	if (verbose)
	{
		printf("%s  (", errs ? " FAIL <---" : "");
		printf("Fref ");
		gsc_label_long_comma(fref);
		printf(", Ndiv ");
		gsc_label_long_comma(ndiv_best);
		printf(", Fsamp %.3f", fsamp_best);
		printf(")\n");
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	ao8c500k_fsamp_get
*
*	Purpose:
*
*		Determine the output sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		verbose	Work verbosely?
*
*		get		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao8c500k_fsamp_get(int fd, int verbose, s32* get)
{
	s32		burst		= -1;
	s32		enable		= -1;
	s32		enable_a	= -1;
	int		errs		= 0;
	s32		fref;
	s32		ndiv_a		= -1;
	double	rate;
	s32		set;
	s32		src			= -1;

	errs	+= ao8c500k_query(fd, -1, 0, AO8C500K_QUERY_FREF_DEFAULT, &fref);

	errs	+= ao8c500k_ioctl_dsl(fd, AO8C500K_IOCTL_CLOCK_ENABLE, &enable);
	errs	+= ao8c500k_ioctl_dsl(fd, AO8C500K_IOCTL_CLOCK_SRC, &src);
	errs	+= ao8c500k_ioctl_dsl(fd, AO8C500K_IOCTL_BURST_ENABLE, &burst);
	errs	+= ao8c500k_ioctl_dsl(fd, AO8C500K_IOCTL_GEN_A_ENABLE, &enable_a);
	errs	+= ao8c500k_ioctl_dsl(fd, AO8C500K_IOCTL_GEN_A_NDIV, &ndiv_a);

	if (errs)
	{
		set	= 0;

		if (verbose)
			printf("FAIL <---  (%d. service errors)\n", __LINE__);
	}
	else if (enable == AO8C500K_CLOCK_ENABLE_NO)
	{
		set	= 0;

		if (verbose)
			printf("0  (%d. Output is disabled.)\n", __LINE__);
	}
	else if (burst == AO8C500K_BURST_ENABLE_YES)
	{
		set	= 0;

		if (verbose)
			printf("Unknown  (%d. Bursting is enabled.)\n", __LINE__);
	}
	else if (src == AO8C500K_CLOCK_SRC_EXT)
	{
		set	= 0;

		if (verbose)
			printf("Unknown  (%d. The clock source is external.)\n", __LINE__);
	}
	else if (enable_a == AO8C500K_GEN_ENABLE_NO)
	{
		set	= 0;

		if (verbose)
			printf("0  (%d. Rate-A Generator is disabled.)\n", __LINE__);
	}
	else
	{
		set	= fref / ndiv_a;

		if (verbose)
		{
			gsc_label_long_comma(set);
			printf(" S/S  (Ref = ");
			gsc_label_long_comma(fref);
			printf(", Source = Rate-A Gen");
			printf(", Ndiv = ");
			gsc_label_long_comma(ndiv_a);
			rate	= ((double) fref) / ndiv_a;
			printf(", Fsamp = %.3f)\n", (double) rate);
		}
	}

	if (get)
		get[0]	= set;

	return(errs);
}



/******************************************************************************
*
*	Function:	ao8c500k_fsamp_ao_report
*
*	Purpose:
*
*		Determine and report the output sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		chan	The index of the channel of interest.
*
*		get		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao8c500k_fsamp_ao_report(int fd, int index, int verbose, int chan, s32* get)
{
	s32		bit			= 0x1L << chan;
	char	buf[64];
	int		errs;
	s32		select		= -1;
	s32		set			= 0;

	if (verbose)
	{
		sprintf(buf, "AO Channel %d", chan);
		gsc_label_index(buf, index);
	}

	errs	= ao8c500k_ioctl_dsl(fd, AO8C500K_IOCTL_CHANNEL_SEL, &select);

	if (select & bit)
	{
		// The channel is enabled.
		errs	+= ao8c500k_fsamp_get(fd, verbose, &set);
	}
	else
	{
		// The channel is disabled.
		set	= 0;

		if (verbose)
			printf("0 S/S  (Channel Disabled)\n");
	}

	if (get)
		get[0]	= set;

	return(errs);
}



/******************************************************************************
*
*	Function:	ao8c500k_fsamp_ao_report_all
*
*	Purpose:
*
*		Determine and report the output sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	Report the overall sample rate here.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao8c500k_fsamp_ao_report_all(int fd, int index, int verbose, s32* fsamp)
{
	s32	chans;
	int	chan;
	int	errs;
	s32	sps;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("AO Sample Rates", index);

	errs	= ao8c500k_query(fd, -1, 0, AO8C500K_QUERY_CHANNEL_QTY, &chans);

	if (errs)
	{
		if (verbose)
			printf("FAIL <---\n");
	}
	else
	{
		if (verbose)
		{
			printf("\n");
			gsc_label_level_inc();
		}

		for (chan = 0; chan < chans; chan++)
		{
			sps		= 0;
			errs	+= ao8c500k_fsamp_ao_report(fd, -1, verbose, chan, &sps);
			total	+= sps;
		}

		if (verbose)
		{
			gsc_label("Overall AO Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");
			gsc_label_level_dec();
		}
	}

	if (fsamp)
		fsamp[0]	= total;

	return(errs);
}


